;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  dosinit.asm
;* Purpose: Initialize the DOS simulator, setup all interupt vectors etc.
;* Entries: _init_minidos, dosfatal, call21
;*
;**************************************************************************
;*
;* Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\..\headers\asm\macros.inc
include ..\..\headers\asm\layout.inc
include ..\..\headers\asm\memory.inc
include ..\public\minidos.inc
include .\dospriv.inc


;
;**************************************************************************
;
; The DOS simulator needs it's own stack when int21h is called. This stack
; is placed in the BSS segment. Note that the real DOS uses three different
; stacks. This will not be simulated.
;
bss_start				; start BSS segment

STACKSIZE	equ	512		; pretty small, so be careful.
STACKSIG	equ	0AA55h		; signature for checking stack overflow

dos_sig		dw	?
dos_stack	db	STACKSIZE dup (?)
dos_tos		label	byte

bss_end


;
;**************************************************************************
;
; And now some variables for our mini-DOS simulator:
;
data_start				; start data segment

		extrn	curr_psp:word	; segment of current psp
		extrn	first_mcb:word	; segment of first MCB in list

		public	old_stack, old_ds, dos_active

old_ds		dw	0		; caller's data segment
old_ax		dw	0		; caller's AX register
old_bx		dw	0		; caller's BX register
old_stack	dd	0		; pointer to caller's stack
dos_active	db	0		; DOS active counter

data_end


;
;**************************************************************************
;
; Define macros to produce jump table entries.
;
tab	macro	funcno

	db	0&funcno&h			;; save function number
if 0&funcno&h ne 0FFh				;; 0FFh means end of table
	dw	offset cgroup:dos&funcno	;; save pointer to label

	extrn	dos&funcno:near			;; define label as near external
endif

	endm

TABSIZE	equ	3			; number of bytes per table entry


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	_init_minidos		; define entry points
	public	call21
	public	dosfatal

; General library functions:

	extrn	_fatal:near
	extrn	prnstr:near


;
;**************************************************************************
;
; Interrupt handler table.
;
inttab	dw	offset cgroup:int20		; terminate program
	dw	offset cgroup:int21		; dos function scheduler
	dw	offset cgroup:unused		; called when prog terminated
	dw	offset cgroup:unused		; called when ctrl-C
	dw	offset cgroup:unused		; called when critical error
	dw	offset cgroup:unused		; direct read from disk
	dw	offset cgroup:unused		; direct write to disk
	dw	offset cgroup:int27		; terminate and stay resident

FRSTINT	equ	20h
INTNUM	equ	($ - inttab) / 2		; number of interrupts


;
;**************************************************************************
;
; Initialize mini-DOS simulator module.
; Input:  1. Arg on stack  -  far pointer to first usable memory address
; Output: None
;
_init_minidos	proc	near

	penter				; setup standard stack frame
	mov	dos_sig,STACKSIG	; setup sig for stack overflow check

; Initialize DOS memory by setting up the first memory control block.

	getarg	ax,0			; get first argument: offset
	getarg	bx,1			; get first argument: segment
	shift	shr,ax,4		; adjust pointer to nearest paragraph
	add	ax,bx			; boundary
	add	ax,2			; one additional paragraph for safety
	mov	bx,LOWMEM
	sub	bx,ax			; check that we have enough memory
	jc	short init3
	cmp	bx,RAMSIZE / 16
	ja	short init4
init3:	mov	ax,FATAL_NOMEM		; fatal error -> not enough memory
	jmp	short dosfatal

init4:	dec	bx			; MCB requires one paragraph of free
	push	ds			; memory
	mov	first_mcb,ax		; save pointer to first MCB
	mov	ds,ax
	mov	ds:[mcb_id],MCB_END_ID
	mov	ds:[mcb_owner],0	; set MCB of first free memory block
	mov	ds:[mcb_size],bx
	pop	ds

; Install interrupt handlers.

	cli
	cld
	push	es
	push	di
	xor	ax,ax
	mov	es,ax
	mov	di,FRSTINT * 4		; first interrupt is 20h
	mov	bx,offset cgroup:inttab	; pointer to interrupt table
	mov	cx,INTNUM
init5:	mov	ax,cs:[bx]
	stosw				; save offset
	mov	ax,cs			; then segment
	stosw
	inc	bx			; continue with next interrupt
	inc	bx
	loop	init5
init6:	pop	di
	pop	es			; that's it
	sti

	pleave
	ret

_init_minidos	endp


;
;**************************************************************************
;
; Handle fatal errors by printing a message and then rebooting.
; Input:  AX  -  error number
; Output: routine does not return
;
dosfatal	proc	near

	sti
	push	ax
	mov	bx,offset cgroup:errmsg
	call	prnstr			; print starting message
	pop	si
	shl	si,1
	mov	bx,cs:errtab[si]	; print error message
	call	prnstr
	jmp	_fatal

; Error messages

errmsg	db	0Dh,0Ah
	db	'DOS ERROR: '
	db	0

; The following table has to be identical with the error numbers in
; the dospriv include file:

errtab	dw	offset cgroup:errnom	; not enough memory
	dw	offset cgroup:errstk	; stack overflow
	dw	offset cgroup:errrty	; reentry
	dw	offset cgroup:errnop	; no active process
	dw	offset cgroup:errmem	; memory corrupt

errnom	db	'no free mem',0
errstk	db	'stack overflow',0
errrty	db	'reentrant call',0
errnop	db	'inv term',0
errmem	db	'mem corrupt',0

dosfatal	endp


;
;**************************************************************************
;
; Jump table for main dispatcher interrupt. Put it into code segment
; (instead const segment) to save space in lower ram area.
;
jmptab	label	byte

; functions for basic I/O:
	irp	x,<01,02,03,04,05,06,07,08,09,0A,0B,0C>
	tab	<&x>
	endm

; functions for system management:
	irp	x,<0D,0E,19,1A,25,2E,2F,30,33,34,35,37,50,51,54,62>
	tab	<&x>
	endm

; functions for time operations:
ifdef NEEDTIME
	irp	x,<2A,2B,2C,2D>
	tab	<&x>
	endm
endif

; functions for memory and process management:
	irp	x,<00,26,31,48,49,4A,4B,4C,4D>
	tab	<&x>
	endm

; functions for file handling:
	irp	x,<3C,3D,3E,3F,40>
	tab	<&x>
	endm

	tab	<FF>			; end of table

; Those functions which were only implemented before DOS2.0 return
; a different error code than those which were introduced with
; DOS2.0 and later. The first "later" function is 30h:

DOS20FN	equ	30h


;
;**************************************************************************
;
; Routine for main dispatcher interrupt 21h. This involves changing to a
; new stack, and then searching for the required function in the function
; table. The function handler has to get the caller's data segment from
; old_ds if necessary.
; Input:  AH  -  DOS function number
;         others depending on function number
; Output: depending on function number
;
int21		proc	far

int21s:	cli
	push	ax
	mov	ax,sp
	add	ax,2
	push	ax
	push	ds			; switch to a new stack without
	mov	ax,NEWDATA		; changing any register
	mov	ds,ax
	pop	old_ds
	pop	word ptr [old_stack + 0]
	mov	word ptr [old_stack + 2],ss
	pop	old_ax
	mov	old_bx,bx
	mov	ss,ax			; set new stack
	mov	sp,offset dgroup:dos_tos

; The DOS interrupt handler is not reentrant, so check for this.

	mov	ax,FATAL_REENTRY
	test	dos_active,0FFh
ifdef IS386
	jnz	dosfatal
else
	jz	short int21r
	jmp	dosfatal
endif

; Search for the function in the function table.

int21r:	mov	bx,offset cgroup:jmptab
	mov	ax,old_ax
int21l:	mov	al,cs:[bx]
	cmp	al,0FFh			; check for end of table
	je	short notfnd
	cmp	al,ah			; found function value?
	je	short int21f
	add	bx,TABSIZE		; not found, loop to next
	jmp	short int21l		; table entry

; Found table entry, so prepare registers and jump to the routine. This
; is done by putting it's address onto the stack, and then executing a
; near return instruction.

int21f:	inc	dos_active		; increment active counter
ifdef IS186
	push	offset cgroup:int21t	; push return address onto stack
else
	mov	ax,offset cgroup:int21t	; push return address onto stack
	push	ax
endif
	push	cs:[bx+1]		; push jump address onto stack
	mov	bx,old_bx
	mov	ax,old_ax		; restore all registers except DS
	sti
	retn				; jump to handler routine

; Return to the caller by restoring the stack without changing any
; registers. Also check for stack overflow. Take care so that the
; flags don't get changed.

int21t:	cli
	pushf
	dec	dos_active		; decrement active counter
	cmp	dos_sig,STACKSIG	; check for stack overflow
	je	short noerr
	mov	ax,FATAL_STACK		; call fatal error handler
	jmp	dosfatal

noerr:	mov	old_bx,bx
int21e:	mov	old_ax,ax
	pop	ax			; get flags
	mov	ss,word ptr [old_stack + 2]
	mov	sp,word ptr [old_stack + 0]
	mov	bx,sp
	mov	byte ptr ss:[bx+4],al	; this is the place where the INT
	mov	bx,old_bx		; operation did put the flags
	mov	ax,old_ax
	mov	ds,old_ds		; restore caller's registers
	iret

; We couldn't find the requested function in the function table, so return
; with an error code in AX and the carry flag set. However, if it is a
; FCB related function, return with the proper error code for that.

notfnd:	cmp	ah,DOS20FN		; check if it's an old function
	jae	short notfn1
	mov	al,0FFh			; return with proper error code
	jmp	short notfn2

notfn1:	mov	ax,ERR_NOFN		; return error code
notfn2:	stc				; set carry flag to indicate error
	pushf
	jmp	short int21e

int21		endp


;
;**************************************************************************
;
; INT20h handler: terminate current program.
; Input:  none
; Output: routine does not return
;
int20		proc	far

	xor	ah,ah
	jmp	int21s			; just call DOS function 00h

int20		endp


;
;**************************************************************************
;
; INT27h handler: terminate current program and stay resident.
; Input:  DX  -  number of bytes in program memory block
; Output: routine does not return
;
int27		proc	far

	add	dx,000Fh
	rcr	dx,1			; round byte count to nearest
	shift	shr,dx,3		; paragraph
	mov	ah,31h			; just call DOS function 31h
	jmp	int21s

int27		endp


;
;**************************************************************************
;
; Handler for unused interrupts
; Input:  none
; Output: none
; Registers changed: none
;
unused		proc	far

	iret

unused		endp


;
;**************************************************************************
;
; Another way for a program to call the function dispatcher is to call
; address psp:0005h. This will jump to the following address.
;
call21		proc	far

	cli
	push	bp
	mov	bp,sp
	push	ax
	mov	ax,[bp + 6]		; get return address to caller
	mov	[bp + 2],ax		; and put it into place
	pushf
	pop	[bp + 6]		; replace the old return address
	pop	ax			; with the flags. this makes the
	pop	bp			; call look like an INT.
	jmp	int21s

call21		endp

;
;**************************************************************************
;
text_end

	end

