/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  ip.c
 * Purpose: Internet Protocol handling, includes ICMP handling
 * Entries: init_ip, ip_reg_proto, ip_send
 *
 **************************************************************************
 *
 * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../../headers/general.h"
#include "../public/packet.h"
#include "../public/net.h"
#include "../public/romlib.h"
#include "./netpriv.h"
#include "./icmp.h"
#include "./arp.h"
#include "./ip.h"


/*
 **************************************************************************
 * 
 * Definition of data structure used for the higher protocol without
 * ICMP. Note that we only support one protocol at a time (which is
 * likely to be UDP). This should be sufficient for a boot rom.
 */
struct s_proto {
	unsigned short  proto;		/* protocol ID			*/
	proto_recv     *phandler;	/* receive handler		*/
};



/*
 **************************************************************************
 * 
 * Global variables:
 */
static struct s_proto pdef;		/* protocol definition		*/
static int pktmtu;			/* MTU as given by packet drvr	*/
static int packetid = 1;		/* packet ID counter		*/



/*
 **************************************************************************
 * 
 * Register a new protocol
 */
int ip_reg_proto(unsigned short proto, proto_recv *handler)
{
  pdef.proto = proto;
  pdef.phandler = handler;
  return(pktmtu);
}



/*
 **************************************************************************
 * 
 * Send an IP packet.
 */
static void ip_send_pkt(char *buf, int size, t_ipaddr dest, unsigned char *addr,
							unsigned short proto)
{
#define ipp ((struct iphdr *)buf)

  /* Prepare IP header */
  ipp->ip_ihl_version = ((IP_VERSION << 4) | ((IP_MIN_HSIZE/4) & 0x0f)) & 0xff;
  ipp->ip_tos         = IP_STD_TOS;
  ipp->ip_tot_len     = htons(IP_MIN_HSIZE + size);
  ipp->ip_id          = htons(packetid++);
  ipp->ip_frag_off    = 0;
  ipp->ip_ttl         = IP_TTL;
  ipp->ip_protocol    = proto;
  ipp->ip_check       = 0;
  ipp->ip_daddr       = dest;
  ipp->ip_saddr       = htonl(myipaddr);
  ipp->ip_check       = ipchksum((unsigned char *)buf, IP_MIN_HSIZE);

  (void)write_packet(buf, IP_MIN_HSIZE + size, htons(ETH_P_IP), addr);

#undef ipp
}



#ifndef NOICMP
/*
 **************************************************************************
 * 
 * ICMP receiver routine
 */
static void icmp_recv(char *buf, unsigned char *addr)
{
#define ipp ((struct iphdr *)buf)
#define icmpp ((struct icmphdr *)(buf + IP_MIN_SIZE))

  int icmplen;

  /*
   * The only ICMP message type we can handle is an ECHO request.
   */
  icmplen = ntohs(ipp->ip_tot_len) - IP_MIN_HSIZE;
  if (icmpp->icmp_type == ICMP_ECHO_REQ) {
	icmpp->icmp_type = ICMP_ECHO_REP;
	icmpp->icmp_check = 0;
	icmpp->icmp_check = ipchksum((unsigned char *)icmpp, icmplen);
	ip_send_pkt(buf, icmplen, ipp->ip_saddr, addr, IPPROTO_ICMP);
  }

#undef icmpp
#undef ipp
}
#endif



/*
 **************************************************************************
 * 
 * IP receiver routine
 */
static void ip_recv(char *buf, int bufsize, unsigned char *addr)
{
#define ipp ((struct iphdr *)buf)

  int size;
  t_ipaddr dadr;

#ifdef DEBUG
  printf("IP: got %d bytes from %x:%x:%x:%x:%x:%x\n",
	bufsize, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
#endif

  /*
   * Do some checks on the received packet. It must be an IP packet
   * with a valid version number and header size. Also the checksum
   * of the header must be correct, and the packet should not be frag-
   * mented.
   */
  size = ntohs(ipp->ip_tot_len);
  if (bufsize <= IP_MIN_HSIZE ||
    ipchksum((unsigned char *)ipp, IP_MIN_HSIZE) != 0 ||
    (ipp->ip_ihl_version & 0xff) != ((IP_VERSION << 4) | (IP_MIN_HSIZE >> 2)) ||
    (ntohs(ipp->ip_frag_off) & 0x3fff))
	return;

  /*
   * The destination address of the packet must be valid. It must either
   * be 0.0.0.0 (which means "any"), 255.255.255.255 for broadcast packets,
   * or our own IP number.
   */
  dadr = ntohl(ipp->ip_daddr);
  if (dadr != IP_ANY && dadr != IP_BROADCAST && dadr != myipaddr &&
      myipaddr != IP_ANY)
	return;

#ifdef DEBUG
  printf("IP: got packet from %lx\n", dadr);
#endif

  /*
   * Handle ICMP packets
   */
#ifndef NOICMP
  if (ipp->ip_protocol == IPPROTO_ICMP) {
	icmp_recv(buf, addr);
	return;
  }
#endif

  /* Update the ARP table using the received packet */
  addcache(addr, ntohl(ipp->ip_saddr));

  /*
   * OK. Now we know, that the IP packet is valid and indeed meant for us.
   * Check if it's for a supported protocol.
   */
  if (pdef.proto != ipp->ip_protocol || pdef.phandler == NULL)
	return;

  /*
   * Send the full packet (including IP header) up one layer. Note that
   * the upper layer routine is not allowed to change anything within
   * the data packet.
   */
  pdef.phandler(buf, size);

#undef ipp
}



/*
 **************************************************************************
 *
 * Send an IP packet.
 */
int ip_send(char *buf, int size, t_ipaddr dest)
{
  unsigned char *addr;

  /* Find a routing to the destination host */
  if ((addr = resolve(dest)) == NULL)
	return(0);

#ifdef DEBUG
  printf("IP: sending %d bytes to %x:%x:%x:%x:%x:%x\n",
	size, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
#endif

  /* And finally send the packet */
  ip_send_pkt(buf, size, htonl(dest), addr, pdef.proto);

  return(size);
}



/*
 **************************************************************************
 * 
 * Network intialization
 */
int init_ip(void)
{
  /* Set module name for error handling */
  module_name = "ip";

  /* Register IP packet type and set the MTU for the interface */
  if (reg_type(htons(ETH_P_IP), ip_recv) != 0)
	return(FALSE);
  if ((pktmtu = get_mtu()) < IP_RECV_SIZE)
	return(FALSE);

  return(TRUE);
}

