/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  udp.c
 * Purpose: UDP handling
 * Entries: init_udp, udp_open, udp_close, udp_read, udp_write
 *
 **************************************************************************
 *
 * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../../headers/general.h"
#include "../public/packet.h"
#include "../public/net.h"
#include "../public/romlib.h"
#include "./netpriv.h"
#include "./ip.h"
#include "./icmp.h"
#include "./udp.h"


/*
 **************************************************************************
 * 
 * Structure used for holding UDP socket information
 */
struct udp_sock {
	unsigned short source;		/* Source port number		*/
	unsigned short dest;		/* Destination port number	*/
	t_ipaddr       daddr;		/* Destination IP address	*/
	char          *buf;		/* pointer to receive buffer	*/
	int            buflen;		/* length of contents in buffer	*/
};


/*
 **************************************************************************
 * 
 * Global variables:
 */
static struct udp_sock usock;		/* socket			*/
static int recv_buf_size;		/* size of receive buffer	*/
static char *writebuf;			/* output buffer		*/



/*
 **************************************************************************
 * 
 * Open a new UDP socket.
 */
void udp_open(t_ipaddr daddr, int source, int dest)
{
  usock.source = source;
  usock.dest = dest;
  usock.daddr = daddr;
}



/*
 **************************************************************************
 * 
 * UDP receiver routine. This gets called by the IP layer when a packet
 * arrives. It returns IP_KEEP if the packet cannot get discarded.
 */
static void udp_recv(char *buf, int size)
{
#define udpp ((struct udphdr *)(buf + IP_MIN_HSIZE))

  /*
   * Check if we have a suitable socket for this packet. Also check if
   * the length of the packet is valid, and if the receive buffer is
   * free. Otherwise discard the packet.
   */
  if (usock.source != ntohs(udpp->udp_dest) || usock.buflen > 0 ||
      (sizeof(struct udphdr) + IP_MIN_HSIZE) > size ||
      (ntohs(udpp->udp_len) + IP_MIN_HSIZE) > size)
	return;

#ifdef DEBUG
  printf("UDP: got packet from port %d to port %d\n",
				ntohs(udpp->udp_source), ntohs(udpp->udp_dest));
#endif

  /*
   * Update the sockets destination port number and get the packet from
   * the IP layer.
   */
  usock.dest = ntohs(udpp->udp_source);
  usock.buflen = ntohs(udpp->udp_len) - sizeof(struct udphdr);
  memcpy(usock.buf, (char *)udpp + sizeof(struct udphdr), usock.buflen);

#undef udpp
}



/*
 **************************************************************************
 * 
 * Read one packet from a UDP socket
 */
int udp_read(char *buf, int bufsize, int timeout, char abortch)
{
  int len;

  /* Wait until we get something */
  set_timeout(timeout);
  while (!chk_timeout() && usock.buflen == 0)
	if (abortch && chkkey() == abortch)
		return(-1);
  if ((len = usock.buflen) > bufsize)
	len = bufsize;
  if (len > 0) {
	memcpy(buf, usock.buf, len);
	usock.buflen = 0;
  }
  return(len);
}



/*
 **************************************************************************
 * 
 * Write one packet to a UDP socket
 */
int udp_write(char *buf, int len)
{
#define udpp ((struct udphdr *)(writebuf + IP_MIN_HSIZE))

#ifdef DEBUG
  printf("UDP: sending from %ld.%d to %ld.%d\n",
			myipaddr, usock.source, usock.daddr, usock.dest);
#endif

  /*
   * Copy the packet into the output buffer. It's size should not be larger
   * than maximum UDP packet
   */
  if (len > UDP_SEND_SIZE)
	len = UDP_SEND_SIZE;
  memcpy(((char *)udpp) + sizeof(struct udphdr), buf, len);
  len += sizeof(struct udphdr);

  /* Prepare UDP datagram and send it out */
  udpp->udp_source = htons(usock.source);
  udpp->udp_dest = htons(usock.dest);
  udpp->udp_len = htons(len);
  udpp->udp_check = 0;
  return(ip_send(writebuf, len, usock.daddr));

#undef udpp
}



/*
 **************************************************************************
 * 
 * Network intialization:
 */
int init_udp(void)
{
  char *cp;

  /* Set module name for error handling */
  module_name = "udp";

  /*
   * Register the UDP protocol with the IP layer and allocate memory for the
   * receive and send buffers.
   */
  recv_buf_size = ip_reg_proto(IPPROTO_UDP, udp_recv);
  if ((cp = malloc(IP_SEND_SIZE + recv_buf_size)) == NULL)
	return(FALSE);
  usock.buf = cp;
  usock.buflen = 0;
  writebuf = cp + recv_buf_size;

  return(TRUE);
}

