;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  pktinit.asm
;* Purpose: Load and initialize the packet driver
;* Entries: _init_packet
;*
;**************************************************************************
;*
;* Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\..\headers\asm\macros.inc
include ..\..\headers\asm\layout.inc
include ..\..\headers\asm\memory.inc
include ..\public\minidos.inc
include .\pktpriv.inc


;
;**************************************************************************
;
; Miscelleneous equates:
;
MIN_INT		equ	60h		; lowest usable packet driver interrupt
MAX_INT		equ	80h		; highest usable packet driver interrupt
DRIVER_ID	equ	<'PKT DRVR'>	; ID at beginning of packet driver
DRIVER_LEN	equ	8		; length of ID string
DRIVER_OFS	equ	3		; offset of ID string


;
;**************************************************************************
;
; Data segment:
;
data_start

		extrn		bufsize:word
		extrn		readbuf:word
		extrn		writebuf:word

data_end


;
;**************************************************************************
;
; BSS segment. This segment gets initialized to zero by the startup code.
;
bss_start

pktoldint	dw		?		; packet driver interrupt no.
progpar		prog_param	<>		; structure for prog params

bss_end


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	_init_packet		; define entry points

; General library functions:

	extrn	prnstr:near
	extrn	prnbyte:near
	extrn	prnword:near
	extrn	_malloc:near
	extrn	_fatal:near


;
;**************************************************************************
;
; Initialize packet driver interface.
; Input:  1. Arg on stack  -  far pointer to packet driver code
;         2. Arg on stack  -  far pointer to driver command line
; Output: None
;
_init_packet	proc	near

	penter				; setup standard stack frame

; Setup program parameter array. Note that we don't have an environment
; or FCB's, so these pointer have to be 0. This will be accomplished
; automatically by the startup code since the progpar structure is in
; the BSS segment.

ifdef IS386
	getarg	eax,2			; get far pointer to command line
	mov	progpar.prog_cmdl,eax	; from stack
else
	getarg	ax,2			; get far pointer to command line
	getarg	cx,3			; from stack
	mov	word ptr [progpar.prog_cmdl+0],ax
	mov	word ptr [progpar.prog_cmdl+2],cx
endif

; Now call the packet driver.

	push	ds
	push	es
	push	si
	push	di
	push	bp
	mov	ax,ds
	mov	es,ax
	mov	bx,offset dgroup:progpar
	getarg	dx,0			; get address to binary image from
	getarg	ds,1			; stack
	mov	ax,4B00h
	int	21h			; call packet driver
	pop	bp
	pop	di
	pop	si
	pop	es
	pop	ds
	jnc	short init2		; check if error

	mov	dx,offset cgroup:dosmsg	; print DOS error message
	xor	bx,bx
	jmp	short initF		; fatal error - reboot system

; Check the return code and issue a message in case of error.

init2:	mov	ah,4Dh
	int	21h			; get packet driver exit code
	cmp	ah,3
	je	short init3
	cmp	al,RET_NORMAL
	jne	short init3
	mov	al,RET_E_UNKNOWN
init3:	cmp	al,RET_NORMAL
	je	short init6
	xor	bx,bx
	xor	ah,ah
	mov	dx,offset cgroup:errmsg	; print error messages
	jmp	short initF

; Packet driver has been loaded. Now determine the driver's interrupt,
; and redirect it to our own interrupt. Using an interrupt which is known
; at assembly time allows the use of the int instruction.

init6:	call	pktfind			; try to find packet driver int
	or	ax,ax
	jnz	short init7
	mov	bx,offset cgroup:intmsg	; print interrupt error message
initF:	jmp	pktfatal

init7:	mov	pktoldint,ax		; save packet driver interrupt
	push	es
	xor	ax,ax
	mov	es,ax			; save packet driver interrupt vector
ifdef IS386
	mov	eax,es:[bx]
	mov	es:[PKTINT * 4],eax
else
	mov	ax,es:[bx+0]
	mov	word ptr es:[PKTINT * 4 + 0],ax
	mov	ax,es:[bx+2]
	mov	word ptr es:[PKTINT * 4 + 2],ax
endif
	pop	es

; Setup the packet driver interface and tell the user that we found something.

	call	pktsetup
	push	ax
	mov	bx,offset cgroup:fndmsg
	call	prnstr
	mov	ax,pktoldint
	call	prnbyte
	mov	bx,offset cgroup:clsmsg
	call	prnstr
	pop	ax
	call	prnbyte
	mov	bx,offset cgroup:crlf
	call	prnstr
	pleave
	ret

; Error messages:

dosmsg	db	'DOS error code: ',0
errmsg	db	'Exit code: ',0
intmsg	db	'No int',0

; Message to tell user about packet driver interrupt:

fndmsg	db	'Found driver at int ',0
clsmsg	db	', class ',0
crlf	db	0Dh,0Ah,0

_init_packet	endp


;
;**************************************************************************
;
; Find packet driver interrupt
; Input:  none
; Output: AX  -  Interrupt number, or 0 if not found
;         BX  -  pointer to interrupt vector in segment 0
; Registers changed: AX, BX, CX
;
pktfind		proc	near

	cld
	push	ds
	push	es
	push	si
	push	di
	mov	ax,cs
	mov	ds,ax
	mov	bx,MIN_INT * 4		; start looking at interrupt 60h
pktf1:	xor	ax,ax
	mov	es,ax
	les	di,es:[bx]		; get interrupt vector
	mov	ax,es
	or	ax,di			; ignore null pointers
	jz	short pktf2
	add	di,DRIVER_OFS		; compute address of ID string
	mov	si,offset cgroup:pktdrv_id
	mov	cx,DRIVER_LEN
	repe	cmpsb			; compare both strings
	jz	short pktf3		; found it
pktf2:	add	bx,4			; advance to next interrupt vector
	cmp	bx,MAX_INT * 4		; at the end of the list?
	jb	short pktf1
	xor	ax,ax
	jmp	short pktf4

pktf3:	mov	ax,bx
ifdef IS186
	shr	ax,2			; return interrupt number
else
	shr	ax,1
	shr	ax,1			; return interrupt number
endif
pktf4:	pop	di
	pop	si
	pop	es
	pop	ds
	ret

; ID string at the beginning of the packet driver code:

pktdrv_id	db	DRIVER_ID

pktfind		endp


;
;**************************************************************************
;
; Initialize the packet driver interface, after the packet driver has been
; loaded. This basically involves getting some info from the driver, and
; then allocating enough memory to hold the receive buffers.
; Input:  none
; Output: AL  -  packet driver class
; Registers changed: AX, BX, CX, DX
;
pktsetup	proc	near

	push	es
	push	di
	mov	ah,GET_PARAMETERS
	int	PKTINT			; get parameters from packet
	mov	ax,ETH_FRAME_MAX	; driver. if driver is too old,
	jc	short pkts3		; assume default value
	mov	ax,es:[di + 4]		; get MTU from driver
	cmp	ax,ETH_FRAME_MAX
	jae	short pkts2
pkts1:	mov	bx,offset cgroup:pktmsg	; invalid packet driver
	jmp	short pktfatal

pkts2:	mov	dl,es:[di + 3]		; get size of MAC address
	cmp	dl,ETH_ALEN		; check if ethernet address length
	jne	short pkts1		; is correct
pkts3:	pop	di
	pop	es
	add	ax,ETH_FLEN		; add FCS size (just for safety)
	mov	bufsize,ax
	shl	ax,1			; compute size of read and write buffers
	add	ax,10h			; one paragraph for extra safety
	push	ax
	call	_malloc			; allocate buffer memory
	pop	dx
	or	ax,ax			; error?
	jnz	short pkts5
pkts4:	mov	bx,offset cgroup:bufmsg
	jmp	short pktfatal		; cannot continue without read buffer

pkts5:	mov	readbuf,ax		; save address of read buffer
	add	ax,bufsize		; compute address of write buffer
	mov	writebuf,ax

	push	ds
	push	si
	mov	ax,(DRIVER_INFO shl 8) + 0FFh
	xor	bx,bx			; dummy handle
	int	PKTINT			; get packet driver info
	pop	si
	pop	ds
	jc	short pkts1		; packet driver is too old
	cmp	ch,CL_DEFAULT		; packet driver is not for ethernet
	jne	short pkts1
	mov	al,ch
	ret

; Error messages:

bufmsg	db	'Cannot alloc buffer',0
pktmsg	db	'Invalid PKTDRVR',0

pktsetup	endp


;
;**************************************************************************
;
; Print fatal error message and exit
; Input:  BX  -  pointer to error message
;         AX  -  output value (only if BX == 0)
;         DX  -  pointer to error message (only if BX == 0)
; Output: routine does not return
;
pktfatal	proc	near

	push	ax
	push	bx
	mov	bx,offset cgroup:fatmsg
	call	prnstr
	pop	bx
	pop	ax
	or	bx,bx
	jnz	short fat1

	push	ax
	mov	bx,dx
	call	prnstr
	pop	ax
	call	prnword
	jmp	short fat2
fat1:	call	prnstr
fat2:	jmp	_fatal

; Preceding error message:

fatmsg	db	0Dh,0Ah
	db	'PKTDRV ERROR: '
	db	0

pktfatal	endp


;
;**************************************************************************
;
text_end

	end

