/*
 **************************************************************************
 *
 * Utility program to create the configuration shell script.
 *
 * Module:  config.c
 * Purpose: Create configuration shell script for bootrom
 * Entries: main
 *
 **************************************************************************
 *
 * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#ifdef MSDOS
#include <io.h>
#include <dir.h>
#endif
#ifdef UNIX
#include <dirent.h>
#endif
#include "../../headers/general.h"


#ifndef ROOTDIR
#define ROOTDIR		"."
#endif
#define KERNELDIR	ROOTDIR "/binaries"
#define PKTDRVDIR	ROOTDIR "/pktdrvr"
#define DEFAULTINT	0x62

#ifdef UNIX
#define DEFAULT_SCRIPT	"mkimage.sh"
#endif
#ifdef MSDOS
#define DEFAULT_SCRIPT	"mkimage.bat"
#endif



/* Public variables */
static char *kernelname = NULL;
static char *pktdrvname = NULL;
static char *pktdrvoptions = NULL;



/* Directory entry name list */
static struct direntry {
	struct direntry *next;
	char             fname[1];
} *dirlist = NULL;



/* List of packet driver options for selected Crynwr packet drivers */
#define HW_IRQ		0x01
#define IO_ADDR		0x02
#define BASE_MEM	0x04
#define AUI_TYPE	0x08
#define DMA_NUM		0x10

static struct pktopt {
	char *pktname;
	int   pktopt;
} pktlist[] = {
	{ "3c501",	HW_IRQ | IO_ADDR },
	{ "3c503",	HW_IRQ | IO_ADDR | AUI_TYPE },
	{ "3c509",	0 },
	{ "arcether",	HW_IRQ | IO_ADDR | BASE_MEM },
	{ "depca",	HW_IRQ | IO_ADDR | BASE_MEM },
	{ "ne1000",	HW_IRQ | IO_ADDR },
	{ "ne2000",	HW_IRQ | IO_ADDR },
	{ "ne2100",	HW_IRQ | IO_ADDR | DMA_NUM },
	{ "smc_wd",	HW_IRQ | IO_ADDR | BASE_MEM },
	{ NULL,		0 }};



/**************************************************************************

		MSDOS dependent functions

 **************************************************************************/


#ifdef MSDOS
/*
 * Convert a unix path to a DOS path with backslashes
 */
static void convpath(char *str)
{
  while (*str) {
	if (*str == '/')
		*str = '\\';
	str++;
  }
}



/*
 * Get a list of directory entries with a specified name pattern.
 */
static int getdirlist(char *pattern)
{
  struct ffblk ff;
  struct direntry *dp;
  int len, count;
  char *src, *dest;

  convpath(pattern);
  if (findfirst(pattern, &ff, FILENAME | EXTENSION | WILDCARDS) < 0)
	return(0);
  count = 0;
  while (TRUE) {
	len = strlen(ff.ff_name) + sizeof(struct direntry);
	if ((dp = (struct direntry *)malloc(len)) == NULL) {
		fprintf(stderr, "not enough memory for directory list\n");
		exit(1);
	}
	for (src = ff.ff_name, dest = dp->fname; *src; src++, dest++) {
		if (*src == '.') break;
		*dest = tolower(*src);
	}
	*dest = '\0';
	dp->next = dirlist;
	dirlist = dp;
	count++;
	if (findnext(&ff) < 0)
		break;
  }
  return(count);
}



/*
 * Create the configuration shell script
 */
static void buildconfig(char *outfname)
{
  FILE *outfile;

  if ((outfile = fopen(outfname, "wt")) == NULL) {
	fprintf(stderr, "cannot create output file %s\n", outfname);
	exit(1);
  }
  fprintf(outfile, "@echo off\n");
  fprintf(outfile, ".\\utils\\pass1 -k %s -p %s -a \"%s\" -o image.bin\n",
					kernelname, pktdrvname, pktdrvoptions);
  fprintf(outfile, ".\\utils\\compress -b15 image.bin\n");
  fprintf(outfile, ".\\utils\\pass2 -i image.biz -l .\\binaries\\floppy.bin -o image.flo\n");
  fprintf(outfile, ".\\utils\\pass2 -i image.biz -l .\\binaries\\rom.bin -o image.rom\n");
  fprintf(outfile, "if exist image.biz del image.biz\n");
  fprintf(outfile, "if exist image.bin del image.bin\n");
  fclose(outfile);
}
#endif



/**************************************************************************

		UNIX dependent functions

 **************************************************************************/

#ifdef UNIX
/*
 * Compare a directory entry against a pattern. This covers wildcards
 * in the same simple way as MSDOS does.
 */
static int dircmp(char *pattern, char *name)
{
  char *cp;

  while (*pattern && *name) {
	if (*pattern == '?' || *pattern == *name) {
		pattern++;
		name++;
	} else {
		if (*pattern != '*')
			break;
		pattern++;
		while (*name && *name != *pattern)
			name++;
	}
  }
  return(!*pattern && !*name);
}



/*
 * Get a list of directory entries with a specified name pattern.
 */
static int getdirlist(char *pattern)
{
  DIR *dir;
  struct dirent *ent;
  struct direntry *dp;
  int len, count;
  char *dirname;
  char *src, *dest;

  /* First open the directory */
  if ((dirname = strdup(pattern)) == NULL) {
	fprintf(stderr, "not enough memory to save directory pattern\n");
	exit(1);
  }
  if ((src = strrchr(dirname, '/')) == NULL) {
	pattern = dirname;
	dirname = ".";
  } else {
	*src++ = '\0';
	pattern = src;
  }
  if ((dir = opendir(dirname)) == NULL) {
	fprintf(stderr, "cannot open directory %s\n", dirname);
	exit(1);
  }

  /* Next step through all directory entries */
  count = 0;
  while ((ent = readdir(dir)) != NULL) {
	if (!dircmp(pattern, ent->d_name))
		continue;
	len = strlen(ent->d_name) + sizeof(struct direntry);
	if ((dp = (struct direntry *)malloc(len)) == NULL) {
		fprintf(stderr, "not enough memory for directory list\n");
		exit(1);
	}
	for (src = ent->d_name, dest = dp->fname; *src; src++, dest++) {
		if (*src == '.') break;
		*dest = tolower(*src);
	}
	*dest = '\0';
	dp->next = dirlist;
	dirlist = dp;
	count++;
  }
  closedir(dir);
  return(count);
}



/*
 * Create the configuration shell script
 */
static void buildconfig(char *outfname)
{
  FILE *outfile;

  if ((outfile = fopen(outfname, "w")) == NULL) {
	fprintf(stderr, "cannot create output file %s\n", outfname);
	exit(1);
  }
  fprintf(outfile, "#!/bin/sh\n");
  fprintf(outfile, "./utils/pass1 -k %s -p %s -a \"%s\" -o image.bin\n",
					kernelname, pktdrvname, pktdrvoptions);
  fprintf(outfile, "compress -b15 image.bin\n");
  fprintf(outfile, "./utils/pass2 -i image.bin.Z -l ./binaries/floppy.bin -o image.flo\n");
  fprintf(outfile, "./utils/pass2 -i image.bin.Z -l ./binaries/rom.bin -o image.rom\n");
  fprintf(outfile, "rm -f image.bin image.bin.Z\n");
  fclose(outfile);
}
#endif



/**************************************************************************

			OS independent functions

 **************************************************************************/


/*
 * Delete directory list
 */
static void deldirlist(void)
{
  struct direntry *dp1, *dp2;

  dp1 = dirlist;
  while (dp1 != NULL) {
	dp2 = dp1;
	dp1 = dp1->next;
	free(dp2);
  }
  dirlist = NULL;
}



/*
 * Get a numeric response from the user
 */
static long getnum(char *prompt, long min, long max, int hex)
{
  long j;

  j = min - 1;
  while (j < min || j > max) {
	printf("  %s (%s): ", prompt, hex ? "hex" : "decimal");
	if ((hex ? scanf("%lx", &j) : scanf("%ld", &j)) != 1) {
		scanf("%*s");
		j = min - 1;
	}
	if (j < min || j > max)
		if (hex)
			printf("  Only allowed values are [0x%lx..0x%lx]\n",
								min, max);
		else
			printf("  Only allowed values are [%ld..%ld]\n",
								min, max);
  }
  return(j);
}



/*
 * Get packet driver options
 */
static void getpktopt(char *pktname)
{
  int i;
  long j;
  char *cp;

  i = 0;
  while (pktlist[i].pktname != NULL) {
	if (!strcmp(pktlist[i].pktname, pktname))
		break;
	i++;
  }
  if (pktlist[i].pktname == NULL)
	return;
  if ((pktdrvoptions = malloc(1024)) == NULL) {
	fprintf(stderr, "cannot allocate memory for pktdrv options\n");
	exit(1);
  }
  sprintf(pktdrvoptions, "0x%x", DEFAULTINT);
  if (pktlist[i].pktopt) {
	cp = pktdrvoptions + strlen(pktdrvoptions);
	*cp++ = ' ';
	printf("Enter packet driver command line options:\n");
	if (pktlist[i].pktopt & HW_IRQ) {
		j = getnum("Hardware IRQ number", 2, 15, FALSE);
		sprintf(cp, "%ld", j);
		cp = pktdrvoptions + strlen(pktdrvoptions);
		if (pktlist[i].pktopt &= ~HW_IRQ)
			*cp++ = ' ';
	}
	if (pktlist[i].pktopt & IO_ADDR) {
		j = getnum("I/O address", 0, 0x1fff, TRUE);
		sprintf(cp, "0x%lx", j);
		cp = pktdrvoptions + strlen(pktdrvoptions);
		if (pktlist[i].pktopt &= ~IO_ADDR)
			*cp++ = ' ';
	}
	if (pktlist[i].pktopt & BASE_MEM) {
		j = getnum("Shared memory address", 0, 0xffff, TRUE);
		sprintf(cp, "0x%lx", j);
		cp = pktdrvoptions + strlen(pktdrvoptions);
		if (pktlist[i].pktopt &= ~BASE_MEM)
			*cp++ = ' ';
	}
	if (pktlist[i].pktopt & DMA_NUM) {
		j = getnum("DMA number", 0, 7, FALSE);
		sprintf(cp, "%ld", j);
		cp = pktdrvoptions + strlen(pktdrvoptions);
		if (pktlist[i].pktopt &= ~DMA_NUM)
			*cp++ = ' ';
	}
	if (pktlist[i].pktopt & AUI_TYPE) {
		j = getnum("AUI type (-1 for default)", -1, 1, FALSE);
		if (j > -1) {
			sprintf(cp, "%ld", j);
			cp = pktdrvoptions + strlen(pktdrvoptions);
		}
	}
  }
  return;
}



/*
 * Let the user select a packet driver.
 */
static void getpktdrv(void)
{
  int num, sel, i;
  struct direntry *dp;
  char buf[1024];

  if ((num = getdirlist(PKTDRVDIR "/*.com")) == 0) {
	fprintf(stderr, "Oops, no packet drivers found!\n");
	exit(1);
  }
  printf("The following %d packet drivers are available:\n", num);
  printf("(0)    user defined packet driver\n");
  for (dp = dirlist, i = 1; dp != NULL; dp = dp->next, i++)
	printf("(%d)    %s\n", i, dp->fname);
  sel = -1;
  while (sel < 0 || sel > num) {
	printf("Select the packet driver you wish to use: ");
	if (scanf("%d", &sel) != 1)
		scanf("%*s");
  }
  if (sel == 0) {
	printf("Enter the full path name of the packet driver: ");
	scanf("%s", buf);
	if ((pktdrvname = malloc(strlen(buf) + 2)) == NULL) {
		fprintf(stderr, "not enough memory for packet driver path\n");
		exit(1);
	}
	strcpy(pktdrvname, buf);
  } else {
	for (dp = dirlist, i = 1; dp != NULL && sel != i; dp = dp->next, i++) ;
	if (dp != NULL) {
		i = strlen(dp->fname) + strlen(PKTDRVDIR) + 5;
		if ((pktdrvname = malloc(i)) == NULL) {
			fprintf(stderr, "not enough memory for packet driver path name\n");
			exit(1);
		}
		strcpy(pktdrvname, PKTDRVDIR "/");
		strcat(pktdrvname, dp->fname);
		strcat(pktdrvname, ".com");
		getpktopt(dp->fname);
	}
  }
  if (pktdrvoptions == NULL) {
	printf("Enter command line arguments for packet driver: ");
	scanf("%s", buf);
	if ((pktdrvoptions = malloc(strlen(buf) + 2)) == NULL) {
		fprintf(stderr, "not enough memory for packet driver options\n");
		exit(1);
	}
	strcpy(pktdrvoptions, buf);
  }
#ifdef MSDOS
  convpath(pktdrvname);
#endif
  deldirlist();
  printf("\n\n\n");
}



/*
 * Let the user select a bootrom kernel.
 */
static void getkernel(void)
{
  int num, sel, i;
  struct direntry *dp;

  if ((num = getdirlist(KERNELDIR "/kernel*.bin")) == 0) {
	fprintf(stderr, "Oops, no kernel binaries found!\n");
	exit(1);
  }
  printf("The following %d kernels are available:\n", num);
  for (dp = dirlist, i = 1; dp != NULL; dp = dp->next, i++) {
	printf("(%d)    %s\t", i, dp->fname);
	if (!strcmp(dp->fname, "kernel"))
		printf("-  standard kernel (recommended)\n");
	else if (!strcmp(dp->fname, "kernelm"))
		printf("-  minimal kernel for small bootroms\n");
	else if (!strcmp(dp->fname, "kernel86"))
		printf("-  kernel for 16-Bit x86 processors\n");
	else
		printf("\n");
  }
  sel = -1;
  while (sel < 1 || sel > num) {
	printf("Select the kernel you wish to use: ");
	if (scanf("%d", &sel) != 1)
		scanf("%*s");
  }
  for (dp = dirlist, i = 1; dp != NULL && sel != i; dp = dp->next, i++) ;
  if (dp != NULL) {
	i = strlen(dp->fname) + strlen(KERNELDIR) + 5;
	if ((kernelname = malloc(i)) == NULL) {
		fprintf(stderr, "not enough memory for kernel path name\n");
		exit(1);
	}
	strcpy(kernelname, KERNELDIR "/");
	strcat(kernelname, dp->fname);
	strcat(kernelname, ".bin");
  }
#ifdef MSDOS
  convpath(kernelname);
#endif
  deldirlist();
  printf("\n\n\n");
}



void main(int argc, char **argv)
{
  char *cp = DEFAULT_SCRIPT;

  if (argc > 1)
	cp = argv[1];

  printf("\nConfiguration program for bootrom, version %d.%d\n\n\n",
							MAJOR_VER, MINOR_VER);
  getkernel();
  getpktdrv();
  printf("Building configuration script for the following parameters:\n");
  printf("Kernel:        \"%s\"\n", kernelname);
  printf("Packet driver: \"%s %s\"\n", pktdrvname, pktdrvoptions);
  buildconfig(cp);
}

