/*
 **************************************************************************
 *
 * Utility program to combine the boot rom kernel with the packet driver
 *
 * Module:  pass1.c
 * Purpose: Combine the bootrom kernel and the packet driver
 * Entries: main
 *
 **************************************************************************
 *
 * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#ifdef MSDOS
#include <io.h>
#endif
#include "../../headers/memory.h"


/* Define for debugging output */
#undef DEBUG


#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif


#define BLKSIZE		512		/* Size of input buffer */
#define MAXARGS		126		/* Maximum size of pktdrvr cmd line */
#define EXESIG		0x5A4D		/* Signature of DOS EXE files */

#define PKTHEADER	4		/* Length of packet driver header */
#define COMSIG		0x4B47		/* Signature for packet driver header */
#define SIZEOFS		2		/* Offset to word with size of pktdrv */


char *progname;
char *krnname = NULL;
char *pktargs = NULL;
char *pktname = NULL;
char *outname = NULL;
int infile;
int outfile;



/*
 * Write one block of data to the output file.
 */
int dowrite(unsigned char *buf, int bufsize)
{
  int ofs;

  if ((ofs = write(outfile, (char *)buf, bufsize)) < 0) {
	perror(outname);
	exit(1);
  }
#ifdef DEBUG
  fprintf(stderr, "w:%d<%s>\n", ofs, outname);
#endif
  return ofs;
}



/*
 * Read one block of data either from the kernel image file or the
 * packet driver binary.
 */
int doread(unsigned char *buf, int bufsize, char *name)
{
  int ofs;

  if ((ofs = read(infile, (char *)buf, bufsize)) < 0) {
	perror(name);
	exit(1);
  }
#ifdef DEBUG
  fprintf(stderr, "r:%d<%s>\n", ofs, name);
#endif
  return ofs;
}



void usage(void)
{
  fprintf(stderr, "usage: %s -k<kernel file> -p<packet driver> ", progname);
  fprintf(stderr, "-a<pktdrv args> [-o<out file>]\n");
  exit(1);
}



main(int argc, char **argv)
{
  unsigned long argofs;
  unsigned long pktofs;
  unsigned long writecnt;
  unsigned short krnlength;
  unsigned char inbuf[BLKSIZE];
  int i, len;
  char *cp;

  /* Determine our own name */
#ifdef MSDOS
  if ((cp = strrchr(argv[0], '\\')) == NULL)
#else
  if ((cp = strrchr(argv[0], '/')) == NULL)
#endif
	cp = argv[0];
  else
	cp++;
  progname = strdup (cp);

  /* Isolate the different command line options */
  for (i = 1; i < argc; i++) {
	if (*argv[i] != '-' || strlen(argv[i]) < 2)
		usage();
	 switch (argv[i][1]) {
		case 'k':  if (strlen(argv[i]) > 2) {
				cp = &argv[i][2];
		           } else {
				if (++i >= argc) usage();
				cp = argv[i];
		           }
		           krnname = strdup(cp);
		           break;
		case 'p':  if (strlen(argv[i]) > 2) {
				cp = &argv[i][2];
		           } else {
				if (++i >= argc) usage();
				cp = argv[i];
		           }
		           pktname = strdup(cp);
		           break;
		case 'o':  if (strlen(argv[i]) > 2) {
				cp = &argv[i][2];
		           } else {
				if (++i >= argc) usage();
				cp = argv[i];
		           }
		           outname = strdup(cp);
		           break;
		case 'a':  if (strlen(argv[i]) > 2) {
				cp = &argv[i][2];
		           } else {
				if (++i >= argc) usage();
				cp = argv[i];
		           }
		           pktargs = strdup(cp);
		           break;
		default:   usage();
	}
  }

  /* Check if all necessary command line options are given */
  if (krnname == NULL || pktname == NULL || pktargs == NULL)
	usage();
  if (strlen(pktargs) > MAXARGS) {
	fprintf(stderr, "%s: packet driver argument too long\n", progname);
	exit(1);
  }

  /* Open the output file */
  if (outname == NULL) {
	outfile = fileno(stdout);
  } else {
#ifdef MSDOS
	if ((outfile = open(outname, O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, 0666)) < 0) {
#else
	if ((outfile = open(outname, O_CREAT | O_WRONLY | O_TRUNC, 0666)) < 0) {
#endif
		perror(outname);
		exit(1);
	}
  }

  /*
   * Open the input file, read the first block, and check type of file.
   * If the input file is a boot floppy image, copy the first block directly
   * to the output file, because it is the floppy boot sector.
   */
  len = BLKSIZE;
#ifdef MSDOS
  if ((infile = open(krnname, O_RDONLY | O_BINARY)) < 0) {
#else
  if ((infile = open(krnname, O_RDONLY)) < 0) {
#endif
	perror(krnname);
	exit(1);
  }

  /*
   * Read the first block and determine the actual size of the rom kernel
   * without the BSS segment.
   */
  if (doread(inbuf, len, krnname) != len) {
	fprintf(stderr, "%s: unexpected end of file\n", krnname);
	exit(1);
  }
  krnlength = *((unsigned short *)(&inbuf[(int)DATAOFS + 4])) + 2;

  /*
   * Now copy the rom image file into the output file. While writing to
   * the output file, count the number of bytes written, so that the offset
   * to the packet driver can be determined. Also only copy the relevant
   * bytes excluding the BSS segment, which has only zero bytes anyway.
   */
  writecnt = 0;
  krnlength -= len;
  while (TRUE) {
	writecnt += dowrite(inbuf, len);
	if (krnlength <= 0) break;
	if ((len = doread(inbuf, BLKSIZE, krnname)) == 0) break;
	if (len > krnlength)
		len = krnlength;
	krnlength -= len;
  }
  close(infile);
  argofs = writecnt;

  /*
   * Next write the packet driver argument line to the output file. The
   * argument string contains a leading byte which gives the number of
   * characters in the string.
   */
  len = strlen(pktargs);
  inbuf[0] = len & 0xFF;
  strncpy((char *)(&inbuf[1]), pktargs, len);
  inbuf[len+1] = 0;
  len += 2;
  writecnt += dowrite(inbuf, len);
  pktofs = writecnt;

  /*
   * Now copy the packet driver into the output file, but retain some
   * space for the packet driver header.
   */
#ifdef MSDOS
  if ((infile = open(pktname, O_RDONLY | O_BINARY)) < 0) {
#else
  if ((infile = open(pktname, O_RDONLY)) < 0) {
#endif
	perror(pktname);
	exit(1);
  }
  len = doread(&inbuf[PKTHEADER], BLKSIZE - PKTHEADER, pktname) + PKTHEADER;
  if (*((unsigned short *)(&inbuf[PKTHEADER])) == EXESIG) {
	fprintf(stderr, "%s: cannot use EXE type packet driver\n", pktname);
	exit(1);
  }
  writecnt = 0;
  while (TRUE) {
	writecnt += dowrite(inbuf, len);
	if ((len = doread(inbuf, BLKSIZE, pktname)) == 0) break;
  }
  close(infile);
  writecnt -= PKTHEADER;

  /*
   * Finally write the packet driver offsets and the packet driver header
   * into the destination file.
   */
  if (lseek(outfile, (long)DATAOFS, 0) < 0) {
	perror(outname);
	exit(1);
  }
  if (pktofs > 65535L || argofs > 65535L) {
	fprintf(stderr, "Kernel image file >64kB\n");
	exit(1);
  }
  *((unsigned short *)(&inbuf[0])) = (unsigned short)pktofs;
  *((unsigned short *)(&inbuf[2])) = (unsigned short)argofs;
  (void)dowrite(inbuf, 4);

  if (lseek(outfile, (long)pktofs, 0) < 0) {
	perror(outname);
	exit(1);
  }
  *((unsigned short *)(&inbuf[0])) = (unsigned short)COMSIG;
  *((unsigned short *)(&inbuf[SIZEOFS])) = (unsigned short)((writecnt + 15) >> 4);
  (void)dowrite(inbuf, PKTHEADER);

  close(outfile);
  exit(0);
}

