/*
 * mknbi.c  -  MaKe NetBoot Image
 *
 * Version 1.3 <Thu May 16 12:20:00 MET DST 1996>
 *
 * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
 *
 * Added support for preloaded ramdisk image and for loading "big" kernels
 * into extended memory
 *                          1996 Markus Gutschke <gutschk@uni-muenster.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <malloc.h>
#include <netdb.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mknbi.h"


char *progname;
char nfsdir[MAXPATHLEN];	/* Directory to mount as root via NFS */
char kname[MAXPATHLEN];		/* Name of kernel image file */
char outname[MAXPATHLEN];	/* Name of output file */
char rdname[MAXPATHLEN];	/* Name of ramdisk image file */
int kimage;			/* File handle for kernel image */
int outfile;			/* File handle for output file */
int rdimage = -1;		/* File handle for ramdisk image */
int rdlocation = -1;		/* Memory location of ramdisk image */
int bigkernel = 0;		/* Flag wether kernel should be loaded high */
int verbose = 0;		/* Debugging flag */
enum {
	rd_auto = RD_AUTO,	/* Ramdisk load mode */
	rd_eom = RD_EOM,
	rd_fixed = RD_FIXED
} rdmode = rd_auto;

int cur_rec_num = -1;		/* Number of current load record */
struct load_header header;	/* Load header */
struct load_record *cur_rec;	/* Pointer to current load record */

unsigned char copyrec_buf[SECTSIZE];



/*
 * Write a buffer into the output file and update the load record
 */
static void putrec(int recnum, char *src, int size)
{
  long isize;
  char *buf;

  isize = ((size / (SECTSIZE + 1)) + 1) * SECTSIZE;
  if ((buf = malloc(isize)) == NULL) {
	perror(progname);
	exit(1);
  }
  memset(buf, isize, 0);
  memcpy(buf, src, size);
  if (write(outfile, buf, isize) != isize) {
	perror(outname);
	exit(1);
  }
  free(buf);
  cur_rec->ilength += isize;
  cur_rec->mlength += isize;
}


/*
 * Copy a certain number of bytes from the kernel image file into the
 * boot file
 */
static int copyrec(int recnum, int remaining, int image, char *name)
{
  int bytes_read, size, i = 1;

  for (bytes_read = 0; remaining > 0 && i > 0;) {
	size = (remaining > SECTSIZE) ? SECTSIZE : remaining;
	if ((i = read(image, copyrec_buf, size)) < 0) {
		perror(name);
		exit(1);
	}
	if (i) {
		putrec(recnum, copyrec_buf, i);
		bytes_read += i;
		remaining -= i;
	}
  }
  return (bytes_read);
}


/*
 * Initialize a load record
 */
static void initrec(int recnum, int segment, int flags, int vendor_size)
{
  if (++cur_rec_num != recnum) {
	fprintf(stderr, "%s: Internal error: image chunks misordered!\n",
								progname);
	exit(1);
  }

  if (cur_rec_num > 0)
	(unsigned char *)cur_rec += ((cur_rec->rlength << 2) & 0x3c) +
					((cur_rec->rlength >> 2) & 0x3c);

  cur_rec->rlength  = (sizeof(struct load_record) - sizeof(union vendor_data)) >> 2;
  cur_rec->rlength |= ((vendor_size + 3) << 2) & 0xf0;
  cur_rec->rtag1   = recnum + VENDOR_OFFSET;
  cur_rec->rflags  = flags;
  cur_rec->address = (unsigned long) segment << 4;
}


/*
 * Process kernel image file
 */
static void do_kernel(void)
{
  int flags;
  int setup_size;

  /* Process the floppy boot loader code */
  initrec(INITNUM, DEF_INITSEG, 0, 0);
  if (copyrec(INITNUM, INITLSIZE, kimage, kname) != INITLSIZE) {
	fprintf(stderr, "%s: Unexpected end of kernel image file\n", progname);
	exit(1);
  }
  cur_rec->mlength = INITMSIZE;

  /* Determine size of setup code */
  if ((setup_size = (int)copyrec_buf[497] * SECTSIZE) > SETUPLSIZE) {
	fprintf(stderr, "%s: Kernel setup code too large\n", progname);
	exit(1);
  }

  /* Process the setup code */
  initrec(SETUPNUM, DEF_SETUPSEG, 0, 0);
  if (copyrec(SETUPNUM, setup_size, kimage, kname) != setup_size) {
	fprintf(stderr, "%s: Unexpected end of kernel image file\n", progname);
	exit(1);
  }
  cur_rec->mlength = SETUPMSIZE;

  /* Process the kernel */
  flags = rdimage >= 0 ? 0 : FLAG_EOF;
  if (bigkernel) {
	initrec(KERNELNUM, BIG_SYSSEG, flags, 0);
	while (copyrec(KERNELNUM, SECTSIZE, kimage, kname) == SECTSIZE) ;
	if (cur_rec->ilength > BIG_SYSLSIZE) {
		fprintf(stderr, "%s: Kernel image too large\n", progname);
		exit(1);
	}
	cur_rec->mlength = BIG_SYSMSIZE;
  } else {
	initrec(KERNELNUM, DEF_SYSSEG, flags, 0);
	while (copyrec(KERNELNUM, SECTSIZE, kimage, kname) == SECTSIZE) ;
	if (cur_rec->ilength > SYSLSIZE) {
		fprintf(stderr, "%s: Kernel image too large\n", progname);
		exit(1);
	}
	cur_rec->mlength = SYSMSIZE;
  }

  /* Process the ramdisk image */
  if (rdimage >= 0) {
	unsigned long imagestart = cur_rec->address + cur_rec->ilength;

	if (imagestart < 0x100000L)
		imagestart = 0x100000L;
	initrec(RAMDISKNUM, 0, (rdmode == rd_eom ? FLAG_B1 : 0) | FLAG_EOF,
					sizeof(cur_rec->vendor_data.rdflags));
	if (rdmode == rd_auto)
		rdlocation = imagestart;
	cur_rec->vendor_data.rdflags = rdmode;
	while (copyrec(RAMDISKNUM, SECTSIZE, rdimage, rdname) == SECTSIZE) ;
	cur_rec->address = rdmode == rd_eom ? cur_rec->ilength : rdlocation;
  }
}


/*
 * Print the usage
 */
static void usage(void)
{
  fprintf(stderr, "Usage: %s [-b] [-v] [-s <max cmdline size>] [-d rom|kernel|<dir>]\n", progname);
  fprintf(stderr, "		[-i rom|kernel|<client:server:gateway:netmask:hostname>]\n");
  fprintf(stderr, "		[-a <append>] [-r <ramdisk> [-l auto|eom|<addr>]]\n");
  fprintf(stderr, "		[[-k] <kernel>] [[-o] <outfile>]\n");
  exit(1);
}


/*
 * Main program
 */
void main(int argc, char **argv)
{
  char *cmdline = NULL;		/* Command line to pass to the kernel */
  char *append = NULL;		/* String to append to end of command line */
  char *addrs = NULL;		/* String containing various addresses */
  char *cp, *ip;
  int cmdl_msize = CMDLMSIZE;
  int vendor_size;
  int i, len;

  /* Determine my own name for error output */
  if ((cp = strrchr(argv[0], '/')) == NULL)
	progname = argv[0];
  else
	progname = ++cp;

  /* Initialize option argments */
  strcpy(nfsdir, DFLT_DIR);
  strcpy(kname, DFLT_IMAGE);
  strcpy(outname, "");
  strcpy(rdname, "");

  /* Parse options */
  opterr = 0;
  while ((i = getopt(argc, argv, "a:bd:hi:k:o:r:s:v")) != EOF)
	switch (i) {
		case 'b': bigkernel++;
			  break;
		case 's': if ((i = atoi(optarg)) < CMDLMSIZE) {
				fprintf(stderr, "%s: command line size must be > %d bytes\n",
							progname, CMDLMSIZE-1);
				exit(1);
		          }
		          cmdl_msize = i;
		          break;
		case 'd': strncpy(nfsdir, optarg, MAXPATHLEN-1);
		          nfsdir[MAXPATHLEN-1] = '\0';
		          break;
		case 'k': strncpy(kname, optarg, MAXPATHLEN-1);
		          kname[MAXPATHLEN-1] = '\0';
		          break;
		case 'o': strncpy(outname, optarg, MAXPATHLEN-1);
		          outname[MAXPATHLEN-1] = '\0';
		          break;
		case 'r': strncpy(rdname, optarg, MAXPATHLEN-1);
		          rdname[MAXPATHLEN-1] = '\0';
		          break;
		case 'l': if (!strcmp("auto",optarg))
		          	rdmode = rd_auto;
		          else if (!strcmp("eom",optarg))
			  	rdmode = rd_eom;
		          else {
			  	rdmode = rd_fixed;
				rdlocation = strtol(optarg, NULL, 0); }
		          break;
		case 'a': if ((append = strdup(optarg)) == NULL) {
		          	perror(progname);
		          	exit(1);
		          }
		          break;
		case 'i': if ((addrs = strdup(optarg)) == NULL) {
				perror(progname);
				exit(1);
		          }
		          break;
		case 'v': verbose++;
		          break;
		case 'h':
		default:  usage();
	}

  /* Parse additional arguments */
  if (optind < argc) {
	strncpy(kname, argv[optind++], MAXPATHLEN-1);
	kname[MAXPATHLEN-1] = '\0';
  }
  if (optind < argc) {
	strncpy(outname, argv[optind++], MAXPATHLEN-1);
	outname[MAXPATHLEN-1] = '\0';
  }
  if (optind != argc)
	usage();

  /* Parse the IP address option */
  if (addrs != NULL && strcmp(addrs, "rom") && strcmp(addrs, "kernel")) {
	struct hostent *hp;
	char *buf, *bp;

	/* Allocate memory for 5 address strings, each with max 15 chars */
	if ((buf = malloc((MAX_ADDR_SIZE + 1) * 5)) == NULL) {
		perror(progname);
		exit(1);
	}
	i = 0;
	ip = addrs;
	bp = buf;
	while (ip != NULL && *ip) {
		if ((cp = strchr(ip, ':')) != NULL) *cp++ = '\0';
		if (strlen(ip) > 0) {
			if ((hp = gethostbyname(ip)) == NULL) {
				fprintf(stderr, "%s: invalid hostname %s\n", progname, ip);
				exit(1);
			}
			if (hp->h_length != sizeof(struct in_addr)) {
				fprintf(stderr, "%s: invalid host address type\n", progname);
				exit(1);
			}
			/* Result from ntoa is smaller than MAX_ADDR_SIZE */
			strcpy(bp, inet_ntoa(*(struct in_addr *) hp->h_addr));
			bp += strlen(bp);
		}
		ip = cp;
		if (i < 3) *bp++ = ':';
		if (i >= 3) break;
		i++;
	}
	for (; i < 4; i++)
		*bp++ = ':';
	/* Finally copy host name */
	if (ip != NULL) {
		if ((i = strlen(ip)) > MAX_ADDR_SIZE)
			i = MAX_ADDR_SIZE;
		strncpy(bp, ip, i);
		bp += i;
	}
	*bp = '\0';
	free(addrs);
	addrs = buf;
  }
  if (addrs == NULL)
	addrs = DFLT_ADDRS;

  /* Verify validity of ramdisk load information */
  if (rdmode == rd_fixed && rdlocation < 0x100000L) {
	fprintf(stderr, "%s: Ramdisk image load location must be above 1M\n",
		progname);
	exit(1);
  }

  /* Open the input and output files */
  if ((kimage = open(kname, O_RDONLY)) < 0) {
	perror(kname);
	exit(1);
  }
  if (strlen(outname) == 0 || !strcmp(outname, "-")) {
	outfile = STDOUT_FILENO;
	strcpy(outname, "stdout");
  } else if ((outfile = creat(outname, 0664)) < 0) {
	perror(outname);
	exit(1);
  }
  if (*rdname && (rdimage = open(rdname, O_RDONLY)) < 0) {
	perror(rdname);
	exit(1);
  }
  if (verbose) {
	fprintf(stderr, "Kernel image file name  = %s\n", kname);
	fprintf(stderr, "Output file name        = %s\n", outname);
	if (*rdname)
		fprintf(stderr, "Ramdisk image file name = %s\n", rdname);
  }

  /* Construct command line to pass to the kernel */
  len = strlen(DFLT_CMDL) + 1;
  if (append != NULL)
	len += strlen(append) + 1;
  if (*nfsdir)
	len += strlen(NFS_ROOT) + strlen(nfsdir) + 1;
  if (*addrs)
	len += strlen(NFS_ADDRS) + strlen(addrs) + 1;
  if (len > CMDLLSIZE) {
	fprintf(stderr, "%s: Command line too long\n", progname);
	exit(1);
  }
  if ((cmdline = malloc(len + 2)) == NULL) {
	perror(progname);
	exit(1);
  }

  strcpy(cmdline, DFLT_CMDL);
  if (*nfsdir) {
	cp = cmdline + strlen(cmdline);
	sprintf(cp, "%s%s", NFS_ROOT, nfsdir);
  }
  if (*addrs) {
	cp = cmdline + strlen(cmdline);
	sprintf(cp, "%s%s", NFS_ADDRS, addrs);
  }
  if (append != NULL && *append) {
	cp = cmdline + strlen(cmdline);
	sprintf(cp, " %s", append);
  }
  if (strlen(cmdline) > cmdl_msize)
	cmdline[cmdl_msize] = '\0';
  if (verbose)
	fprintf(stderr, "Kernel command line     = \"%s\"\n", cmdline);

  /* Initialize the boot header */
  vendor_size = (sizeof(VENDOR_ID) / sizeof(unsigned long) + 1) * sizeof(unsigned long);
  memset(&header, sizeof(header), 0);
  header.magic           = HEADER_MAGIC;
  header.hlength         = (unsigned char)(((int) &header.dummy - (int) &header) / sizeof(unsigned long)) & 0x0f;
  header.hlength        |= (unsigned char)((vendor_size / sizeof(unsigned long)) << 4) & 0xf0;
  header.locn.segment    = DEF_HEADERSEG;
  header.locn.offset     = 0;
  header.execute.segment = DEF_BOOTLSEG;
  header.execute.offset  = 0;
  header.bootsig         = BOOT_SIGNATURE;
  memcpy(&header.dummy, VENDOR_ID, sizeof(VENDOR_ID));
  if (write(outfile, &header, sizeof(header)) < 0) {
	perror(outname);
	exit (1);
  }

  /* Initialize pointer to first load record */
  cur_rec = (struct load_record *)&(header.dummy[vendor_size]);

  /* Process the boot loader record */
  if (first_data_size > BOOTLLSIZE) {
	fprintf(stderr, "%s: Boot loader too large\n", progname);
	exit(1);
  }
  initrec(BOOTLNUM, DEF_BOOTLSEG, 0, 0);
  putrec(BOOTLNUM, (char *) &first_data, first_data_size);
  cur_rec->mlength = BOOTLMSIZE;

  /* Process the command line */
  initrec(CMDLNUM, DEF_CMDLSEG, 0, 0);
  putrec(CMDLNUM, cmdline, strlen(cmdline) + 1);
  cur_rec->mlength = cmdl_msize;

  /* Process the kernel image */
  do_kernel();

  /* After writing out all these stuff, finally update the boot header */
  if (lseek(outfile, 0, 0) != 0) {
	perror(outname);
	exit(1);
  }
  if (write(outfile, &header, sizeof(header)) < 0) {
	perror(outname);
	exit (1);
  }
  exit(0);
}
